/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.core.attribute;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.regex.Pattern;

import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeDB.SortingMethod;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.VirtualNode;

public class Attribute implements Comparable<Attribute> {
	private UUID origin;
    private String key;
    private Object value, newvalue;
    private AttributeType type; 
    private AttributeGenerator generator = null;
    private static Set<String> statusNames = new HashSet<String>(),
    							descNames = new HashSet<String>();
    {
    	registerDescriptionName(TreeNode.ATTR_DESCRIPTION);
    	registerStatusName(TreeNode.ATTR_TYPE);
    }
    public static void registerStatusName(String name)
    {
    	statusNames.add(name);
    }
    public static void registerDescriptionName(String name)
    {
    	descNames.add(name);
    }
	public enum ParameterAvailability {
		LOCAL, DIRECT_CHILDREN, SUBTREE
	};
	private ParameterAvailability availability = ParameterAvailability.LOCAL;
	public static boolean affectsPosition(TreeNode node, String attr){
		SortingMethod sm = node.getTreeDB().getSortingMehod();
		if(sm == SortingMethod.BY_INDEX || sm==SortingMethod.BY_LOCATIONS && 
				node.getTreeDB().getTreeLogic().usesUndexByDefault(node.getType()))
			return attr == TreeDB.INDEX_ATTRIBUTE || affectsAttribute(node, attr, TreeDB.INDEX_ATTRIBUTE);
		if(sm == SortingMethod.BY_NAMES || sm == SortingMethod.BY_LOCATIONS)
			return affectsId(node, attr);
		if(attr.equals(VirtualNode.ATTR_HIDDEN_STATUS))
			return true;
		return node.getAttribute(Requirement.ATTR_LOCATIONS)!=null && attr.equals(Requirement.ATTR_LOCATIONS) || 
				affectsId(node, attr);
	}
	public static boolean affectsId(TreeNode node, String attr){
		if(attr.equals("_id")&&node.getName()==null)
			return true;
		if(attr.equals(TreeNode.ATTR_NAME))
			return true;
		return affectsAttribute(node, attr, TreeNode.ATTR_NAME);
	}
	public static boolean affectsAttribute(TreeNode node, String what, String target){
		if(what.equals(target))
			return true;
		if(node.getAttribute(target)!=null)
		{
			for(String s:node.getRevertDependencies(target))
				if(affectsAttribute(node, what, s))
					return true;
		}
		return false;
	}
	public static boolean affectsStatus(TreeNode node, String attr){
		if(statusNames.contains(attr))
			return true;
		if(VirtualNode.ATTR_TARGET.equals(attr))
			return true;
		if(node.getTreeDB().getDefReferenceName().equals(attr))
			return true;
		if(node.getAttribute(attr)!=null && node.getAttribute(attr).getRawType()==AttributeType.REFERENCE)
			return true;
		for(String desc:statusNames)
			if(node.getAttribute(desc)!=null)
			{
				for(String s:node.getRevertDependencies(desc))
					if(affectsAttribute(node, attr, s))
						return true;
			}
		return false;
	}
	public static boolean affectsDescription(TreeNode node, String attr){
		if(descNames.contains(attr))
			return true;
		for(String desc:descNames)
			if(node.getAttribute(desc)!=null)
			{
				for(String s:node.getRevertDependencies(desc))
					if(affectsAttribute(node, attr, s))
						return true;
			}
		return false;
	}
	public ParameterAvailability getAvailability() {
		return availability;
	}

	public static String getRegExp(String varname) {
		return "\\{"
				+ varname
				+ "\\}";
	}
	public boolean moreVisibleThan(Attribute rag2) {
		if (rag2 == null)
			return true;
		ParameterAvailability pa1 = this.getAvailability(), pa2 = rag2
				.getAvailability();
		/*int q1=getMaster().getQualifiedId().length(),q2=rag2.getMaster().getQualifiedId().length();
		if(q1>q2)
			return true;
		if(q2>q1)
			return false;*/
		if(pa1.equals(ParameterAvailability.LOCAL))
			return !pa2.equals(ParameterAvailability.LOCAL);
		else
			if(pa1.equals(ParameterAvailability.DIRECT_CHILDREN))
				return pa2.equals(ParameterAvailability.SUBTREE)||
					   pa2.equals(ParameterAvailability.DIRECT_CHILDREN);
			else
				return pa2.equals(ParameterAvailability.SUBTREE);
	}

	
	public void setAvailability(ParameterAvailability p)
	{
		ParameterAvailability old = availability;
		availability = p;
		if(!p.equals(old)){
			//getMaster().removeIterator(this);
			//getMaster().addExistingAttributeIterator(this);
			//updateMaster();
		}
	}
	//assume that all the rest are updated 
	public boolean isAvaiableOn(TreeNode node) {
		//if it is local then available
		if(getOrigin().equals(node.getUUId()))
			return true;
		//if not it can be in parent
		TreeNode par = node.getParent(true);
		Attribute a = null;
		if(par!=null)
			a = par.getAttribute(getKey(),getOrigin());
		//null =  var seems to be deleted or not from parent
		if(a != null){
			//local vars are not visible...
			if(a.getOrigin().equals(par) && a.getAvailability().equals(ParameterAvailability.LOCAL))
				return false;
			if(a.getAvailability().equals(ParameterAvailability.DIRECT_CHILDREN))
			{
				while (par!=null){
					TreeNode p2 = par.getParent(true);
					if(p2!=null && p2.getAttribute(getKey(),getOrigin())!=null)
						return false;
					par = par.getCloneParent();
				}
				//if there is no that parent which parent has this attribute then we in a proper place
				return true;
			}
			//if subtree var is visible then all is correct
			return a.getAvailability().equals(ParameterAvailability.SUBTREE);
		}
		TreeNode cp = node.getCloneParent();
		if(cp!=null)
			if(isAvaiableOn(cp))
				return true;
		//if var is not defined here and not in clones or its parents then it is incorrect
		return false;

	}

    public Attribute(UUID node, AttributeType type, String key, Object value, AttributeGenerator gen) {
    	this(node,type,key,null,value,gen);
    }
    public Attribute(UUID node, AttributeType type, String key, Object newvalue, Object value) {
    	this(node,type,key,newvalue,value,null);
    }
    public Attribute(UUID node, AttributeType type, String key, Object newvalue, Object value,AttributeGenerator gen) {
    	this.origin = node;
    	this.type = type;
        this.key = key.replace((char)160, ' ');
        if(type == AttributeType.STRING || type == AttributeType.REFERENCE){
        	if(value!=null)
        		value = value.toString().replace((char)160, ' ');
        	if(newvalue!=null)
        		newvalue = newvalue.toString().replace((char)160, ' ');
        }
        this.value = value;
        this.newvalue = newvalue;
    	this.generator = gen;
    	if(gen!=null){
    		gen.setAttribute(this);
    		this.type = AttributeType.GENERATED;
    		//TODO send type to generator
    	}
    }
    
    public Attribute(TreeNode node, AttributeType type, String key, Object value) {
    	this(node.getUUId(), type, key, value);
    }
    public Attribute(UUID origin, AttributeType type, String key, Object value) {
        this(origin,type,key,null,value,null);
    }

    public String getKey() {
        return key;
    }

    public <T> T getGRawValue(){
		return (T)value;
    }
    public Object getValue() {
    	Object o = value;
    	if(newvalue != null)
    		o = newvalue;
    	else{
    		if (type == AttributeType.ENUM)
    		{
    			//TODO show * for not in list enum
    			if(value instanceof AppliedEnumValue)
    				return ((AppliedEnumValue)value).getValue();
    			return value.toString();
    		}
    	}
    	if(o == null && getRawType().equals(AttributeType.ENUM))
    		return "";
    	if(o!=null && getResultType().equals(AttributeType.LIST))
    		return toList((ListValue) value);
    	//if(o!=null && getResultType().equals(AttributeType.LIST)&&!(o instanceof ))
    		//return toList((Object[]) o);
        return o;
    }

    private List<Object> toList(ListValue o) {
		// TODO Auto-generated method stub
    	List<Object> ret = new ArrayList<Object>();
    	for(Object i:o.getValues())
    		if(o.getType() == AttributeType.LIST)
    			ret.add(toList((ListValue) i));
    		else
    			ret.add(i);
		return ret;
	}

	public AttributeType getRawType() {
        return type;
    }

    public AttributeType getResultType() {
    	if(type==AttributeType.ENUM)
    		if(value instanceof AppliedEnumValue){
    			return ((AppliedEnumValue)value).getType();
    		}else
    			return AttributeType.STRING;
    	if(type==AttributeType.GENERATED && generator!=null)
    		return generator.getAttributeType();
        return type;
    }

    @Override
    public String toString() {
        return "["+(availability==ParameterAvailability.LOCAL?"L":availability==ParameterAvailability.DIRECT_CHILDREN?"D":"S")+"]'"+key+"':="+(newvalue!=null?newvalue:"null") + "{"+value + "} ("+super.toString()+") G: "+(getGenerator()==null?"none":getGenerator().toString());
    }

    @SuppressWarnings("unchecked")
    @Override
    public boolean equals(Object obj) {
        if(super.equals(obj)) return true;
        if (!(obj instanceof Attribute)||obj==null) {
            return false;
        }
        Attribute attr = (Attribute) obj;
        if(!key.equals(attr.getKey())) {
            return false;
        }
        if(!type.equals(attr.getRawType())) {
            return false;
        }
        if(!availability.equals(attr.getAvailability()))
        	return false;
        
        if (getResultType() == AttributeType.LIST && getValue() instanceof List && attr.getValue() instanceof List) {
            List<Object> list1 = (List<Object>)getValue();
            List<Object> list2 = new ArrayList<Object>((List<Object>)attr.getValue());
            for (Object str : list1) {
                if (list2.contains(str)) {
                    list2.remove(str);
                } else {
                    return false;
                }
            }
            if (!list2.isEmpty()) {
                return false;
            }
        }
        if(type == AttributeType.ENUM && attr.getRawType()== AttributeType.ENUM){
            AppliedEnumValue v1 = getGRawValue(), v2 = attr.getGRawValue();
            if(v1 == null || v2 == null)
            	return false;
            if(!v1.equals(v2))
            	return false;
        }
        if(generator==null && attr.generator!=null || generator!=null && attr.generator==null || attr.generator!=null && generator != null 
        		&& !attr.generator.getUUID().equals(generator.getUUID()))
        	return false;
        
        if(value==null&&attr.getValue()!=null || value!=null && !valEquals(getValue(),attr.getValue())) {
            return false;
        }
        return true;
    }
    public boolean weakEquals(Object obj) {
        if(super.equals(obj)) return true;
        if (!(obj instanceof Attribute)||obj==null) {
            return false;
        }
        Attribute attr = (Attribute) obj;
        if(!key.equals(attr.getKey())) {
            return false;
        }
        if(!this.getResultType().equals(attr.getResultType())) {
            return false;
        }
        if (getResultType() == AttributeType.LIST && getValue() instanceof List && attr.getValue() instanceof List) {
            List<Object> list1 = (List<Object>)getValue();
            List<Object> list2 = new ArrayList<Object>((List<Object>)attr.getValue());
            for (Object str : list1) {
                if (list2.contains(str)) {
                    list2.remove(str);
                } else {
                    return false;
                }
            }
            if (!list2.isEmpty()) {
                return false;
            }
        }
        if(getValue()==null&&attr.getValue()!=null || getValue()!=null && !valEquals(getValue(),attr.getValue())) {
            return false;
        }
        return true;
    }
    private static boolean valEquals(Object value2, Object object) {
    	if (value2 instanceof ListValue && object instanceof ListValue){
    		ListValue o1=(ListValue)value2,o2=(ListValue)object;
    		if(o1.getValues().size() == o2.getValues().size() && o1.getType()==o2.getType()){
    			return Arrays.equals(o1.getValues().toArray(), o2.getValues().toArray());
    		}else
    			return false;
    	}
		return value2.equals(object);
	}
	static final Pattern valid_inner_name_pattern = Pattern.compile("[A-Za-z0-9а-яА-Я_[$].-]+[A-Za-z0-9а-яА-Я._[$] -]*");
    static final Pattern valid_name_pattern = Pattern.compile("[A-Za-z0-9а-яА-Я.-]+[A-Za-z0-9а-яА-Я._[$] -]*");
    public static boolean checkNameIsValid(String name, boolean inner){
        return inner ? valid_inner_name_pattern.matcher(name).matches() : valid_name_pattern.matcher(name).matches();
    }

    @Override
    public int compareTo(Attribute o) {
        return this.getKey().compareTo(o.getKey());
    }

    @SuppressWarnings("unchecked")

    public static Attribute deepCopy(Attribute attr) {
    	return deepCopy(attr,null);
    }
    public static Attribute deepCopy(Attribute attr, UUID newuid) {
    	if(attr == null)
    		return null;
    	Attribute a = null;
        if(a==null)
        	a = new Attribute(newuid!=null?newuid:attr.origin, attr.type, attr.key, null,attr.getRawValue(), attr.generator!=null?attr.generator.copy():null);
        a.setAvailability(attr.getAvailability());
        return a;
    }

    public AttributeType getListType(){
    	if(getRawType().equals(AttributeType.GENERATED)&&getGenerator()!=null)
    		return getGenerator().getAttributeListType();
    	if(getRawType().equals(AttributeType.LIST))
    		return ((ListValue) value).getType();
    	if(getRawType().equals(AttributeType.ENUM_DEFINITION))
    		return((EnumDefinitionValue)value).getType(); 
    	if(getResultType().equals(AttributeType.ENUM))
    		return ((AppliedEnumValue)value).getType();
    	return null;
    }
	public Object getRawValue() {
		return value;
	}

	public UUID getOrigin() {
		return origin;
	}
	
	public boolean isGenerated(){
		return getRawType().equals(AttributeType.GENERATED)&&generator!=null;
	}

	public boolean isMoreVisible(Attribute a, UUID where) {
		if(where.equals(getOrigin())){
			if(!where.equals(a.getOrigin()))
				return true;
		}else{
			if(where.equals(a.getOrigin()))
				return false;
		}
		if(isGenerated()){
			if(a.isGenerated()){
				return this.moreVisibleThan(a);
			}
			else 
				return false;
		}else
			if(a.isGenerated()){
				return true;
			}
			else 
				return this.moreVisibleThan(a);
	}
	public static String valueToString(AttributeType at, Object value){
		StringBuilder sb = new StringBuilder();
		if(!checkValueByType(at, value))
			return "INCORRECT!";
		EnumValue ev = null;
		if(value instanceof EnumValue){
			ev = (EnumValue)value;
			value = ev.getValue();
			sb.append("|");
		}
		if(at == AttributeType.LIST){
			ListValue lv = (ListValue)value;
			AttributeType at2 = lv.getType();
			sb.append(at2);
			sb.append("=[");
			Iterator<Object> it = lv.iterator();
			while(it.hasNext()){
				sb.append(valueToString(at2,it.next()));
				if(it.hasNext())
					sb.append(",");
			}
			sb.append("]");
		}
		else if (at == AttributeType.ENUM_DEFINITION){
			EnumDefinitionValue lv = (EnumDefinitionValue)value;
			AttributeType at2 = lv.getType();
			sb.append(at2);
			sb.append("=[");
			Iterator<EnumValue> it = lv.iterator();
			while(it.hasNext()){
				sb.append(valueToString(at2,it.next().getValue()));
				if(it.hasNext())
					sb.append(",");
			}
			sb.append("]");
		}
		else 
			if(value!=null)
				sb.append(value.toString());
			else
				sb.append("NULL");
		if(ev!=null){
			if(ev.getComment()!=null){
				sb.append("{");
				sb.append(ev.getComment());
				sb.append("}");
			}
			sb.append("|");
		}
		return sb.toString();
	}
	public void setValue(Object object) {
		value = object;
		newvalue = null;
	}
	public void setNewValue(Object object) {
		newvalue = object;
	}
	public AttributeGenerator getGenerator() {
		if(getRawType().equals(AttributeType.GENERATED))
			return generator;
		return null;
	}
	public void setGenerator(AttributeGenerator ag) {
		generator = ag;
	}

	public Attribute copy() {
		return copy(null);
	}
	public Attribute copy(UUID newuid) {
		return Attribute.deepCopy(this,newuid);
	}

	public boolean isValueRaw() {
		return newvalue == null;
	}
	public static boolean checkValueByType(AttributeType at, Object value2) {
		if(at==AttributeType.ENUM_DEFINITION)return value2 instanceof EnumDefinitionValue;
		if(at==AttributeType.ENUM)return value2 instanceof EnumValue;
		if(at==AttributeType.LIST)return value2 instanceof ListValue;
		if(at == AttributeType.BOOL)
			return value2 instanceof Boolean || value2 instanceof String;
		if(at == AttributeType.INT)
			return value2 instanceof Integer || value2 instanceof String;
		if(at == AttributeType.REFERENCE)
			return value2 instanceof String;
		//TODO additional checks
		return true;
	}

}
