/*
 * Copyright 2009-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.editors.browser;

import java.util.List;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.ui.PlatformUI;

import com.unitesk.requality.core.ITreeLogic.IdContainsWrongChars;
import com.unitesk.requality.core.ITreeLogic.IdEmpty;
import com.unitesk.requality.core.ITreeLogic.PastingIntoLeafException;
import com.unitesk.requality.core.ITreeLogic.SameIdExistsInChildren;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeLogic;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.ui.cnf.RequalityCNF;
import com.unitesk.requality.eclipse.ui.dialogs.SelectNodeDialog;
import com.unitesk.requality.eclipse.ui.dialogs.SelectReqDialog;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;

public class AddSelectionToRequirement extends Action {
    protected ReqMarker reqMarker;
    private Requirement target;
    private String nextId;
    private String newName;

    public AddSelectionToRequirement(ReqMarker reqMarker) {
        super("Add to Requirement...");
        this.reqMarker = reqMarker;
    }

    public AddSelectionToRequirement(ReqMarker reqMarker, Requirement target) {
        super("Add to Requirement: " + target.getNameOrId());
        this.reqMarker = reqMarker;
        this.target = target;
    }

    public AddSelectionToRequirement(ReqMarker reqMarker, Requirement target,
            String nextId) {
        super("Create new Requirement: "
                + (!target.getQualifiedId()
                        .equals(Requirement.getTypeRootQId()) ? target
                        .getNameOrId() + "/" : "") + nextId);
        this.reqMarker = reqMarker;
        this.nextId = nextId;
        this.target = target;
    }

    public AddSelectionToRequirement(ReqMarker reqMarker, Requirement req,
            String id, String newName) {
        this(reqMarker, req, id);
        this.newName = newName;
        // TODO: fix hack
        String label = "Create new Requirement: "
                + (!target.getQualifiedId()
                        .equals(Requirement.getTypeRootQId()) ? target
                        .getNameOrId() + "/" : "") + newName;
        setText(label);
    }

    @Override
    public boolean isEnabled() {
        try {
            boolean s = reqMarker.getBrowser().isSelectable();
            boolean b = reqMarker.getBrowser().isRemovable();
            List<TreeNode> selectedRequirements = RequalityCNF.getSelectedNodes();
            if (selectedRequirements != null
                    && (selectedRequirements.size()!=1)) {
                return false;
            }
            Location loc = reqMarker.getSelectedLocation();
            if (loc != null && nextId == null && target != null && target.getLocation(loc.getQualifiedId()) != null){
                return false;
            }
            return s || b;
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
    }

    @Override
    public void run() {
        TreeDB db = reqMarker.getTreeDB();
        //Multi-requirement locations support
        if (reqMarker.getBrowser().isRemovable()) {
            Location loc = reqMarker.getSelectedLocation();
            if (loc != null) {
                Requirement targetReq = createTargetRequirement(loc.getQualifiedId());
                if (targetReq != null && targetReq.getLocation(loc.getQualifiedId()) == null) {
                    if(db.startTransaction("Add Selection to Requirement",true)){
                    	targetReq.addLocation(loc.getQualifiedId());
                    	targetReq.saveAttributes();
                    	db.commit();
                    }
                    reqMarker.fireSelectionChange();
                }
            }
        } else
        //Create new location
        if (reqMarker.getBrowser().isSelectable()) {
        	reqMarker.getBrowser().setEnabled(false);
            Requirement targetReq = createTargetRequirement(null);
            reqMarker.getBrowser().setEnabled(true);
            if (targetReq != null) {
                try {
                    target = targetReq;
                    Object[] jsData;
                    if(db.startTransaction("Add Selection to Requirement",true)){
                    	jsData = reqMarker.addSelectionTo(targetReq, isHidden());
                    	//if location is not added, then rollback
                    	if (jsData == null){
                    		db.rollback();
                    		return;
                    	}
                    	db.commit(); // Add Selection to Requirement
                    	RequalityCNF view = (RequalityCNF) PlatformUI.getWorkbench()
                            .getActiveWorkbenchWindow().getActivePage()
                            .findView(RequalityCNF.ID);
                    	if (view != null) {
                    		view.selectReveal(new StructuredSelection(targetReq));
                    	}	
                    	reqMarker.setTargetLocation((Location) jsData[1]);
                    }
                    reqMarker.fireSelectionChange();
                } catch (Exception e) {
                    e.printStackTrace();
                    db.rollback();
                }
            }
        }
    }

    private boolean isHidden() {
        return (nextId == null) && (newName != null);
    }

    private Requirement createTargetRequirement(final String existingLocationId) {
        if (target == null) {
            SelectNodeDialog reqSelectDlg = new SelectReqDialog(PlatformUI
                    .getWorkbench().getActiveWorkbenchWindow().getShell(),
                    reqMarker.getTreeDB()) {
                        @Override
                        public boolean canSelect() {
                            if (!super.canSelect()) return false;
                            if (!selectedNode.getType().equals(Requirement.TYPE_NAME)) return false;
                            Requirement req = (Requirement) selectedNode;
                            if (req.getLocation(existingLocationId) != null) return false;
                            return true;
                        }};
            List<TreeNode> selectedRequirements = RequalityCNF.getSelectedNodes();
            if (selectedRequirements != null
                    && selectedRequirements.get(0).equals(Requirement.TYPE_NAME)) {
                reqSelectDlg.select(selectedRequirements.get(0));
            }
            reqSelectDlg.setCanSelectRoot(false);
            if (reqSelectDlg.open() == SelectNodeDialog.OK) {
                if (reqSelectDlg.isNewNodeSelected()) {
                }
                Activator.getDefault().setActiveNode(reqSelectDlg.getResult(),
                        reqSelectDlg.getResult().getType());
                return (Requirement) reqSelectDlg.getResult();
            }
            return null;
        }
        if (nextId != null) {
            //add check to possibility of adding
            //new Requirement to target
            try {
                TreeLogic l = target.getTreeDB().getTreeLogic();
                if(l.canBeChildOfExc(Requirement.TYPE_NAME, target) &&
                        l.isChildIdValidExc(target, nextId)){
                    target.getTreeDB().startTransaction("Create Requirement");
                    Requirement req = (Requirement) target.createChild(nextId);
                    if (newName != null) {
                        req.setName(newName);
                        req.saveAttributes();
                    }
                    target.getTreeDB().commit();
                    return req;
                }else{
                    MessageDialog.openError(PlatformUI.getWorkbench()
                            .getActiveWorkbenchWindow().getShell(), "Add Element",
                            "Element of type '" + Requirement.TYPE_NAME 
                            + "' cannot be created in " + target.getUserFriendlyName());
                    return null;
                }
            } catch (PastingIntoLeafException e) {
                MessageDialog.openError(PlatformUI.getWorkbench()
                        .getActiveWorkbenchWindow().getShell(), "Add Element",
                        "Can't add "+Requirement.TYPE_NAME+" to "+target.getTypeLabel()+" with "
                        + "Test Purpose child");
                return null;
        	} catch (IllegalStateException e) {
        		MessageDialog.openError(PlatformUI.getWorkbench()
                    .getActiveWorkbenchWindow().getShell(), "Add Element",
                    "Some actions are already performing");
        		return null;
        	} catch (SameIdExistsInChildren e) {
                MessageDialog.openError(PlatformUI.getWorkbench()
                        .getActiveWorkbenchWindow().getShell(), "Add Element",
                        "Element with same id exists in " + 
                        target.getUserFriendlyName());
                return null;
            } catch (IdEmpty e) {
                MessageDialog.openError(PlatformUI.getWorkbench()
                        .getActiveWorkbenchWindow().getShell(), "Add Element",
                        "Element id is empty");
                return null;
            } catch (IdContainsWrongChars e) {
                MessageDialog.openError(PlatformUI.getWorkbench()
                        .getActiveWorkbenchWindow().getShell(), "Add Element",
                        "Element id contains wrong character - " 
                        + e.getWrongCharacter());
                return null;
            }
        }
        return target;
    }
}
